/******************************************************************************
* To the extent possible under law, Sebastian Roith
* <https://www.roith-unterrichtet.de> has waived all copyright and related or
* neighboring rights to this script. This work is published from Germany.
******************************************************************************/

/**
* Verwaltet eine Liste aus Textbausteinen
*/
function Liste(
    objSpiel,       // Zuständige Reihenfolgenspiel-Instanz
    typeDest,       // Handelt es sich um die Liste, die der Nutzer zusammenstellen soll?
    keyEmpty,       // Schlüssel von objSpiel.data für den anzuzeigenden Text, wenn die Liste leer ist
    arrIndizes      // Array aus Indizes von objSpiel.data.parts für anzuzeigende Listenpunkte
) {
    this.objSpiel = objSpiel;
    this._typeDest = typeDest;
    this.keyEmpty = keyEmpty;
    this.items = [];
    this.ensureList(arrIndizes.length);
    this.addArray(arrIndizes);
    this.ready = true;
}

// Zuständige Reihenfolgenspiel-Instanz
Liste.prototype.objSpiel = null;

// Handelt es sich um die Liste, die der Nutzer zusammenstellen soll?
Liste.prototype._typeDest = false;

// Schlüssel von this.objSpiel.data für den anzuzeigenden Text, wenn die Liste leer ist
Liste.prototype.keyEmpty = 'empty';

// Enthaltene Listenpunkt-Instanzen
Liste.prototype.items = [];

// DOM-Repräsentation der Liste
Liste.prototype.node = null;

// Liste initialisiert?
Liste.prototype.ready = false;

/**
* Prüfen, ob bereits eine Liste angelegt ist
* @return Boolean       Liste vorhanden?
*/
Liste.prototype.isList = function() {
    return (this.node && (['OL', 'UL'].indexOf(this.node.nodeName.toUpperCase()) != -1));
}

/**
* Sicherstellen, dass this.node eine (true) bzw. keine (false) Liste ist
* @return undefined
*/
Liste.prototype.ensureList = function(
    boolList        // Soll eine Liste (true) oder ein Hinweis auf eine leere Liste (false) vorhanden sein?
) {
    if (this.node && (this.isList() == boolList))
        return;
    var node;
    if (boolList)
        node = ScoTools.createHtml(this._typeDest ? 'ol' : 'ul');
    else
        node = ScoTools.createHtml('p', null, { "class": "emptylist" }, this.objSpiel.data[this.keyEmpty] ? this.objSpiel.data[this.keyEmpty] : 'Leere Liste');
    if (this.node)
        this.node.parentNode.replaceChild(node, this.node);
    else
        this.objSpiel.nodeParent.appendChild(node);
    this.node = node;
}

/**
* Einen Listenpunkt erstellen
* @return Listenpunkt       Erstellte Listenpunkt-Instanz
*/
Liste.prototype.add = function(
    index       // Index des hinzuzufügenden Textbausteins aus objSpiel.data.parts
) {
    return new Listenpunkt(this, index);
}

/**
* Mehrere Listenpunkte erstellen
* @return undefined
*/
Liste.prototype.addArray = function(
    arrIndizes      // Array aus Indizes der hinzuzufügenden Textbausteine
) {
    for (var n = 0; n < arrIndizes.length; ++n)
        this.add(arrIndizes[n]);
}

/**
* Einen Listenpunkt aufnehmen und ein Li-Element für ihn erstellen
* @return HTMLElement   Erstelltes LI-Element
*/
Liste.prototype.getSpace = function(
    objListenpunkt      // Listenpunkt, der in Liste aufgenommen werden soll
) {
    this.ensureList(true);
    this.items.push(objListenpunkt);
    if (this.items.length > 1)
        this.items[this.items.length - 2].update();
    this.reportChange();
    return ScoTools.createHtml('li', this.node);
}

/**
* Zum Index index gehörenden Text ermitteln
* @return String        Text
*/
Liste.prototype.getText = function(
    index       // Index des Textbausteins in objSpiel.data.parts
) {
    if (!this.objSpiel.data.parts[index])
        return undefined;
    return this.objSpiel.data.parts[index].text;
}

/**
* Die Indizes der Textbausteine in der Liste ermitteln
* @return Array     Array aller Indizes
*/
Liste.prototype.getIndizes = function() {
    return this.items.map(function(el) { return el.index; });
}

/**
* Darauf reagieren, dass der Listenpunkt objListenpunkt angeklickt wurde
* @return Boolean       Konnte keine Standardaktion durchgeführt werden?
*/
Liste.prototype.procClick = function(
    objListenpunkt      // Angeklickter Listenpunkt
) {
    return this.objSpiel.procClick(objListenpunkt);
}

/**
* Den Listenpunkt objListenpunkt entfernen
* @return undefined
*/
Liste.prototype.remove = function(
    objListenpunkt      // Zu entfernenden Listenpunkt
) {
    Liste.remove(objListenpunkt);
}

/**
* Prüfen, ob diese Liste editierbar ist
* @return String    "never" → Liste ist grundsätzlich nicht editierbar, "no" → Liste aktuell nicht editierbar, "yes" → Liste aktuell editierbar
*/
Liste.prototype.getListEditable = function() {
    return this.objSpiel.getListEditable(this);
}

/**
* Prüfen, ob grundsätzlich Editieren erlaubt ist
* @return String    "no" → Editieren deaktiviert, "yes" → Editieren aktiviert
*/
Liste.prototype.getEditable = function() {
    return this.objSpiel.getListEditable();
}

/**
* Einen Listenpunkt aus seiner Liste entfernen
* @return undefined
*/
Liste.remove = function(
    objListenpunkt      // Zu entfernenden Listenpunkt
) {
    var indexAlt = 0;
    objListenpunkt.liste.items = objListenpunkt.liste.items.filter(function(el, index) {
        if (el == objListenpunkt) {
            indexAlt = index;
            return false;
        } else
            return true;
    });
    if ((indexAlt > 0) && objListenpunkt.liste.items.length)
        objListenpunkt.liste.items[indexAlt - 1].update();
    if (indexAlt < objListenpunkt.liste.items.length)
        objListenpunkt.liste.items[indexAlt].update();
    if (!objListenpunkt.liste.items.length)
        objListenpunkt.liste.ensureList(false);
    if (objListenpunkt.node && objListenpunkt.node.parentNode)
        objListenpunkt.node.parentNode.removeChild(objListenpunkt.node);
    objListenpunkt.liste.reportChange();
}

/**
* Feststellen, ob objListenpunkt der erste Listenpunkt auf seiner Liste ist
* @return Boolean   Erster Listenpunkt?
*/
Liste.isTop = function(
    objListenpunkt      // Fraglicher Listenpunkt
) {
    return (objListenpunkt == objListenpunkt.liste.items[0]);
}

/**
* Feststellen, ob objListenpunkt der letzte Listenpunkt auf seiner Liste ist
* @return Boolean   Erster Listenpunkt?
*/
Liste.isBottom = function(
    objListenpunkt      // Fraglicher Listenpunkt
) {
    return (objListenpunkt == objListenpunkt.liste.items[objListenpunkt.liste.items.length - 1]);
}

/**
* objSpiel von Liste-Instanz informieren, dass der Listenpunkt objListenpunkt entfernt werden soll
*/
Liste.requestRemove = function(
    objListenpunkt      // Fraglicher Listenpunkt
) {
    objListenpunkt.liste.objSpiel.requestRemove(objListenpunkt);
}

/**
* Den Listenpunkt objListenpunkt an die erste Stelle seiner Liste verschieben
* @return undefined
*/
Liste.toTop = function(
    objListenpunkt      // Fraglicher Listenpunkt
) {
    var oldPos = Liste.getPos(objListenpunkt);
    if (!oldPos)
        return;
    var arrTop = [ objListenpunkt.getTop() ];
    var n;
    for (n = 0; n < oldPos; ++n)
        arrTop.push(objListenpunkt.liste.items[n].getTop());
    objListenpunkt.liste.node.insertBefore(objListenpunkt.node, objListenpunkt.liste.node.firstChild);
    objListenpunkt.liste.items = objListenpunkt.liste.items.filter(function(el) { return (el != objListenpunkt) });
    objListenpunkt.liste.items.unshift(objListenpunkt);
    for (n = 0; n < arrTop.length; ++n) {
        objListenpunkt.liste.items[n].update();
        objListenpunkt.liste.items[n].animate(arrTop[n]);
    }
    objListenpunkt.liste.reportChange();
}

/**
* Den Listenpunkt objListenpunkt an die letzte Stelle seiner Liste verschieben
* @return undefined
*/
Liste.toBottom = function(
    objListenpunkt      // Fraglicher Listenpunkt
) {
    var oldPos = Liste.getPos(objListenpunkt);
    var liste = objListenpunkt.liste;
    if (liste.items.length - 1 == oldPos)
        return;
    var arrTop = [];
    var n;
    for (n = oldPos + 1; n < liste.items.length; ++n)
        arrTop.push(liste.items[n].getTop());
    arrTop.push(objListenpunkt.getTop());
    objListenpunkt.liste.node.appendChild(objListenpunkt.node);
    objListenpunkt.liste.items = objListenpunkt.liste.items.filter(function(el) { return (el != objListenpunkt) });
    objListenpunkt.liste.items.push(objListenpunkt);
    for (n = 0; n < arrTop.length; ++n) {
        objListenpunkt.liste.items[oldPos + n].update();
        objListenpunkt.liste.items[oldPos + n].animate(arrTop[n]);
    }
    objListenpunkt.liste.reportChange();
}

/**
* Ermittelt, an wievielter Stelle der Listenpunkt objListenpunkt in seiner Liste steht
* @return Number    Index in der Liste (beginnend bei 0)
*/
Liste.getPos = function(
    objListenpunkt      // Fraglicher Listenpunkt
) {
    return objListenpunkt.liste.items.indexOf(objListenpunkt);
}

/**
* Den Listenpunkt objListenpunkt um eine Stelle früher in der Liste aufführen
* @return undefined
*/
Liste.up = function(
    objListenpunkt      // Fraglicher Listenpunkt
) {
    var n1 = Liste.getPos(objListenpunkt);
    Liste.exchange(objListenpunkt.liste, n1, n1 - 1);
}

/**
* Den Listenpunkt objListenpunkt um eine Stelle später in der Liste aufführen
* @return undefined
*/
Liste.down = function(
    objListenpunkt      // Fraglicher Listenpunkt
) {
    var n1 = Liste.getPos(objListenpunkt);
    Liste.exchange(objListenpunkt.liste, n1, n1 + 1);
}

/**
* Die Listenpunkt n1 und n2 (je Indizes von objListe.items) vertauschen
* @return undefined
*/
Liste.exchange = function(
    objListe,       // Zu bearbeitende Liste
    n1,             // Index dies ersten Listenpunktes in objListe.items
    n2              // Index dies zweiten Listenpunktes in objListe.items
) {
    if ((n1 == n2) || (n1 < 0) || (n2 < 0) || (n1 >= objListe.items.length) || (n2 >= objListe.items.length))
        return;
    if (n2 < n1) {
        var t = n1;
        n1 = n2;
        n2 = t;
    }
    var li1 = objListe.items[n1];
    var li2 = objListe.items[n2];
    objListe.items[n1] = li2;
    objListe.items[n2] = li1;
    var top1 = li1.getTop();
    var top2 = li2.getTop();
    var nextSibling = li2.node.nextSibling;
    objListe.node.insertBefore(li2.node, li1.node);
    objListe.node.insertBefore(li1.node, nextSibling);
    li1.update();
    li2.update();
    li1.animate(top1);
    li2.animate(top2);
    objListe.reportChange();
}

/**
* Anklickbarkeit der Schaltflächen ggf. ändern
* @return undefined
*/
Liste.prototype.update = function(
    n   // Optional: Index der fraglichen Listenpunkt-Instanz (Standardwert: Alle Listenpunkte)
) {
    if (Number.isFinite(n))
        this.items[n].update();
    else {
        for (var z = 0; z < this.items.length; ++z)
            this.items[z].update();
    }
}

/**
* Den n-ten Listenpunkt der Liste als richtig (correctness = true), falsch (correctness = false) oder unbekannt (correctness = undefined) kennzeichnen
* @return undefined
*/
Liste.prototype.markCorrectness = function(
    n,              // Optional: Index der fraglichen Listenpunkt-Instanz  (Standardwert: Alle Listenpunkte)
    correctness     // Ist der Listenpunkt richtig (true), falsch (false) oder noch nicht geprüft (undefined)?
) {
    if (Number.isFinite(n))
        this.items[n].markCorrectness(correctness);
    else
        this.items.map(function(li) { li.markCorrectness(correctness); });
}

/**
* Alle Listenpunkte aus der Liste löschen
* @return undefined
*/
Liste.prototype.clear = function() {
    this.ready = false;
    while (this.items.length)
        this.items[0].remove();
    this.ready = true;
    this.reportChange();
}

/**
* Das Spiel informieren, dass die Liste verändert wurde
* @return undefined
*/
Liste.prototype.reportChange = function() {
    if (this.ready)
        this.objSpiel.procChange(this);
}
